#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import os
import pathlib


class FileTextReadingFailure(Exception):
    pass


def read_file_text(path: str | pathlib.Path | os.PathLike, encoding: str | None = None) -> str:
    if not path:
        raise FileTextReadingFailure(f"Got empty path to read: {path=}")
    try:
        prepared_path = pathlib.Path(path)
    except Exception as e:
        raise FileTextReadingFailure(f"Failed to cast {path=} to pathlib.Path: {e=}") from None

    if not prepared_path.exists():
        raise FileTextReadingFailure(f"No file found by {path=}")
    if not prepared_path.is_file():
        raise FileTextReadingFailure(f"{path=} leads to non file object")
    try:
        return prepared_path.read_text(encoding=encoding)
    except Exception as e:
        raise FileTextReadingFailure(f"Failed to read text from file by {path=}: {e=}") from None



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses

from .._hint import StructureHint, FieldHint
from .._type import get_name_from_type

DEFAULT_REQUIRED_VALUE_PLACEHOLDER = 'REQUIRED'
DEFAULT_DYNAMIC_VALUE_PLACEHOLDER = '...'


def make_ini_from_hint(
        hint: StructureHint,
        *,
        depth: int = 0,
        add_optional: bool = True,
        required_value_placeholder: str = DEFAULT_REQUIRED_VALUE_PLACEHOLDER,
        dynamic_value_placeholder: str = DEFAULT_DYNAMIC_VALUE_PLACEHOLDER,
) -> str:
    section_name_to_lines: dict[str, list[str]] = {}
    section: list[str] | None = None
    for f in hint.fields:
        if isinstance(f, FieldHint):
            section = section or section_name_to_lines.setdefault('', [])
            type_repr = get_name_from_type(f.type)
            if f.default is dataclasses.MISSING:
                section.append(f'; {f.name}: {type_repr}')
                section.append(f'{f.name} = {required_value_placeholder}')
            elif add_optional:
                pass
                if callable(f.default):
                    default_str = f', default={f.default.__module__}.{f.default.__qualname__}'
                    section.append(f'; {f.name} = {dynamic_value_placeholder} ; {type_repr}{default_str}')
                else:
                    section.append(f'; {f.name} = {f.default} ; {type_repr}')
        elif isinstance(f, StructureHint):
            if depth:
                raise ValueError(f"INI format does not support more than one nesting level, got at least two: {f.name}")
            section_name_to_lines[f.name] = [make_ini_from_hint(
                f, depth=depth + 1, add_optional=add_optional, required_value_placeholder=required_value_placeholder,
                dynamic_value_placeholder=dynamic_value_placeholder
            )]
        else:
            raise TypeError(f"unexpected type: {type(f).__qualname__}")

    blocks = []
    if non_section_lines := section_name_to_lines.pop('', []):
        blocks.append('\n'.join(non_section_lines))
    for section_name, lines in section_name_to_lines.items():
        blocks.append(f'[{section_name}]\n' + ('\n'.join(lines)))
    return '\n\n'.join(blocks)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#

from typing import Any


def is_none(val: Any) -> bool:
    return val is None


def is_not_none(val: Any) -> bool:
    return val is not None



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field


@dataclass(frozen=True, slots=True)
class InstanceIdConfig:
    instance_id_template: str = "service:{company}/{group}/{project}, built:{build}, host:{host}, user:{user}, run:{started_at}"
    # instance_id_template: str = "{company}:{group}:{project}::{build}:{environment}({shard_id}){started_at}"
    template_keyword_to_abm_url: dict[str, str] = field(default_factory=lambda: {
        "company": "kwargs://context|_.company",
        "group": "kwargs://context|_.project_group",
        "project": "kwargs://context|_.project_name",
        "build": "import://build_info|_.build_id",
        "environment": "envvar://ENVIRONMENT",
        "shard_id": "envvar://SHARD_ID",
        "started_at":
            "none://|__import__('datetime').datetime.now(__import__('datetime').timezone.utc).astimezone().isoformat()",
        "host": "none://|__import__('socket').gethostname()",
        "user": "none://|__import__('getpass').getuser()",
    })



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import json

from init_helpers._json._alias import JsonLoaded


def read_json_file_by_path(path: str) -> JsonLoaded:
    with open(path) as file:
        return json.load(file)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .get_instance_id import get_instance_id
from .instance_id_config import InstanceIdConfig
from .parsed_abm_url import ParsedAbmUrl



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Any


def append_if(condition: bool, list_: list[Any], *args: Any) -> None:
    """
    if condition is True, append all values to list
    >>> append_if(True, l := [0], 1, 2, 3)
    >>> l
    [0, 1, 2, 3]
    >>> append_if(False, l := [0], 1, 2, 3)
    >>> l
    [0]
    """
    if condition:
        for value in args:
            list_.append(value)



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import string

from .instance_id_config import InstanceIdConfig
from .parsed_abm_url import ParsedAbmUrl


async def get_instance_id(config: InstanceIdConfig, **kwargs) -> str:
    keys_required = {i[1] for i in string.Formatter().parse(config.instance_id_template) if i[1] is not None}
    if missing_keys := keys_required - config.template_keyword_to_abm_url.keys():
        raise KeyError(f'Missing keywords: {missing_keys}') from None
    return config.instance_id_template.format_map({
        key: await ParsedAbmUrl.from_str(config.template_keyword_to_abm_url[key]).get(**kwargs) for key in keys_required
    })



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import logging
import sys
from dataclasses import dataclass
from typing import Callable, Any, Type, TypeVar

from .repr_arg_name_to_exception import repr_arg_name_to_exception
from .._json import read_json_file_by_path
from .._ini import make_ini_from_hint, read_ini_file_by_path
from .._hint import get_hint_from_dataclass
from .._dataclass import FieldErrors, dict_to_dataclass
from .._exception import repr_exception
from .._toml import read_toml_file_by_path

T = TypeVar('T')
logger = logging.getLogger(__file__)


@dataclass
class Arg:
    caster: Callable[[str], Any] = str
    hint: str = None

    def __call__(self, *args, **kwargs):
        return self.caster(*args, **kwargs)

    def __str__(self):
        return self.hint or 'no description'

    @staticmethod
    def json_file(hint: str = None) -> 'Arg':
        return Arg(read_json_file_by_path, hint)

    @staticmethod
    def ini_file_to_dataclass(dataclass_type: Type[T]) -> 'Arg':
        def ini_file_path_to_dataclass(path: str) -> T:
            return dict_to_dataclass(read_ini_file_by_path(path), dataclass_type)
        return Arg(ini_file_path_to_dataclass, hint=make_ini_from_hint(get_hint_from_dataclass(dataclass_type)))

    @staticmethod
    def toml_file_to_dataclass(dataclass_type: Type[T]) -> 'Arg':
        def toml_file_path_to_dataclass(path: str) -> T:
            return dict_to_dataclass(read_toml_file_by_path(path), dataclass_type)
        return Arg(toml_file_path_to_dataclass, hint=make_ini_from_hint(get_hint_from_dataclass(dataclass_type)))

    @staticmethod
    def ini_file(hint: str = None) -> 'Arg':
        return Arg(read_ini_file_by_path, hint)


def parse_args(args: list[str] | None = None, **kwargs: type | Arg) -> list:
    if args is None:
        args = sys.argv[1:]

    usage = f'{sys.argv[0]} {" ".join(kwargs)}'
    full_usage = f'{usage}\nArguments:\n'
    for i, key in enumerate(kwargs):
        full_usage += f'{i+1}. {key}:\n{kwargs[key]}\n'

    arg_name_to_exception = {}
    arg_name_result = {}
    for arg_name, caster in kwargs.items():
        if not args:
            arg_name_to_exception[arg_name] = ValueError("missing argument")
            continue
        arg = args.pop(0)
        try:
            arg_name_result[arg_name] = caster(arg)
        except Exception as e:
            arg_name_to_exception[arg_name] = e

    if arg_name_to_exception:
        logger.warning('Usage:'.upper())
        logger.warning(usage)
        arg_name_to_hint = {
            key: str(type_or_arg) if isinstance(type_or_arg, Arg) else type_or_arg.__qualname__
            for key, type_or_arg in kwargs.items()}
        for line in repr_arg_name_to_exception(arg_name_to_exception, arg_name_to_hint=arg_name_to_hint):
            logger.critical(line)
        exit(2)
    result = [arg_name_result[key] for key in kwargs]
    if len(result) == 1:
        return result[0]
    return result
